 /* Copyright (C) 2008 severalnines.com

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA */


/**
 * The LogEvent Thread handles the log events
 * LogEvents should be either
 *  - ALARMS
 *  - LOG messages
 * Some LogEvents affects the cluster state and what actions we do:
 * Disconnected
 * CommunicationOpened
 * Connected
 * StopForced
 *
 * By looking at the event 
 * StartPhaseCompleted
 * it is possible to record time spent in certain start phases, 
 * but you have to implement the timer.
 *
 * APIs (such as if the mysqlds are connected or not) 
 * can be monitored using the MonitorThread or by looking at the
 * Disconnect/Connect log events. *
 * Please note that the _monitoring thread_ will try to start up
 * failed nodes in NOT_STARED state when the monitor application
 * starts.
 */





#include <time.h>
#include "logevent.h"
#include "connect.h"
#include "statistics.h"
//#include "handler.h"
#include "alarm.h"
#include "log.h"

#include <mgmapi_debug.h>
#include <ndbd_exit_codes.h>

extern configparam_t params;
extern mysql_t mysql;
extern int g_clusterid;

statistics_t  *  perf_stats;
statistics_mem_t  *  mem_stats;
statistics_backup_t  *  backup_stats;
statistics_restore_t  *  restore_stats;


int g_mem_report_cycle=10;

static int filter[] = { 15, NDB_MGM_EVENT_CATEGORY_BACKUP,
			15, NDB_MGM_EVENT_CATEGORY_CONNECTION,
			15, NDB_MGM_EVENT_CATEGORY_NODE_RESTART,
			15, NDB_MGM_EVENT_CATEGORY_STARTUP,
			15, NDB_MGM_EVENT_CATEGORY_STATISTIC,
			15, NDB_MGM_EVENT_CATEGORY_ERROR,
			0 };


void init()
{
  mem_stats = (statistics_mem_t*)malloc(256*sizeof(statistics_mem_t));
  perf_stats = (statistics_t*)malloc(256*sizeof(statistics_t));
  backup_stats = (statistics_backup_t*)malloc(256*sizeof(statistics_backup_t));
  restore_stats = (statistics_restore_t*)malloc(256*sizeof(statistics_restore_t));

  if(mem_stats==0)
    cmon_exit(-1);
  if(perf_stats==0)
    cmon_exit(-1);

  
  memset(perf_stats,0,256*sizeof(statistics_t));      
  memset(mem_stats,0,256*sizeof(statistics_mem_t));      
  memset(backup_stats,0,256*sizeof(statistics_backup_t));      
  memset(restore_stats,0,256*sizeof(statistics_restore_t));      

  char msg[1024];
  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "delete from node_statistics");
  
  executeQuery(msg,&mysql_conn[LOG_THREAD]);


  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "delete from cluster_statistics");
  
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  int retries=0;
  int max_retries=10;
 retry:
  if(retries==max_retries)
    {
      log_message(LOG_ERR, "Could not determine how many data nodes there are in the cluster - file a bug report");
      cmon_exit(-1);
    }
  char sqlbuf[255];
  memset(sqlbuf,0,sizeof(sqlbuf));
  snprintf( sqlbuf,sizeof(sqlbuf),"SELECT nodeid FROM node_state WHERE node_type='NDBD' AND  cid=%d", g_clusterid);
  int a=strlen(sqlbuf);
  MYSQL * mysql = &mysql_conn[LOG_THREAD].mysql;
  if(mysql_real_query(mysql , sqlbuf, a ))
    {
      
      memset(msg,0,sizeof(msg));
      snprintf(msg, sizeof(msg), "Query %s failed: Error: %s",
	       sqlbuf,
	       mysql_error(mysql));
      log_message(LOG_ERR, msg);
      mysql_close(mysql);
      cmon_exit(-1);
    }       
  else // query succeeded, process any data returned by it
    {
      MYSQL_RES * result = mysql_store_result(mysql);
      if (result)  // there are rows
	{
	  MYSQL_ROW row;
	  while ((row = mysql_fetch_row(result)))
	    {
	      
	      memset(msg,0,sizeof(msg));
	      sprintf(msg,
		      "INSERT IGNORE INTO node_statistics("
		      "cid, nodeid" 
		      ") VALUES (%d, %d)",g_clusterid, atoi(row[0]) );  
	      executeQuery(msg,&mysql_conn[LOG_THREAD]);
	    }
	  mysql_free_result(result);
	}
      else
	{
	  retries++;
	  sleep(1);
	  goto retry;
	}
    }

  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "insert into cluster_statistics("
	  "id" 
	  ") values (%d)",g_clusterid );  
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  
}


void logger_create_table (int table_id)
{
  /*
    map table id to name
   */
  printf("table id=%d\n", table_id);
}

void logger_db (int source_node, 
	       int node, 
	       int s, 
	       int loglevel,
	       const char * logevent, 
	       const char * message)
{
  char msg[512];
  char severity[16];
  int  unix_level=0;
  switch(s)
    {            
    case NDB_MGM_EVENT_SEVERITY_DEBUG:
      sprintf(severity,"%s", "DEBUG");
      unix_level=LOG_DEBUG;
      break;
      /** Informational messages*/
      
    case NDB_MGM_EVENT_SEVERITY_INFO:
      sprintf(severity,"%s", "INFO");
      unix_level=LOG_INFO;
      break;
      /** Conditions that are not error condition, but might require handling.
       */
    case NDB_MGM_EVENT_SEVERITY_WARNING:
      sprintf(severity,"%s", "WARNING");
      unix_level=LOG_WARNING;
      break;      
      /** Conditions that, while not fatal, should be corrected. */
    case NDB_MGM_EVENT_SEVERITY_ERROR:
      sprintf(severity,"%s", "ERROR");
      unix_level=LOG_ERR;
      break;            
      /** Critical conditions, like device errors or out of resources */
    case NDB_MGM_EVENT_SEVERITY_CRITICAL:
      sprintf(severity,"%s", "CRITICAL");
      unix_level=LOG_CRIT;
      break;      
	/** A condition that should be corrected immediately,
	 *  such as a corrupted system
	 */
    case NDB_MGM_EVENT_SEVERITY_ALERT:
      sprintf(severity,"%s", "ALERT");            
      unix_level=LOG_ALERT;
      break;      
    default:
      sprintf(severity,"%s", "UNKNOWN");
      break;
      
    }
  
  sprintf(msg,
	  "insert into cluster_log("
	  "cid,"
	  "source_nodeid,"
	  "nodeid,"
	  "severity,"
	  "loglevel,"
	  "event,"
	  "message) values ("
	  "%d,%d,%d,'%s',%d,'%s','%s'"
	  ")", 
	  g_clusterid,
	  source_node,
	  node,
	  severity,
	  loglevel,
	  logevent,
	  message);
  log_message(unix_level, message );
  executeQuery(msg,&mysql_conn[LOG_THREAD]);	         
}


void logger_restore_datalog (int source_node, 
			   int node, 
			   int backupid, 
			   bool log)
{
  char msg[512];  
  sprintf(msg,
	  "update restore "
	  "set "
	  "%s=%llu, " /*records*/
	  "%s=%llu, "  /*bytes*/
	  "status='%s' "
	  "where "
	  "backupid=%d and "
	  "ndb_nodeid=%d and cid=%d",	  
	  log ? "log_records":"records",
	  restore_stats[node].n_records64,
	  log ? "log_bytes ":"bytes ",	      
	  restore_stats[node].n_bytes64,
	  log ? "RestoreLog":"RestoreData",
	  backupid,
	  node,g_clusterid);



  
  log_message(LOG_INFO, msg);
  
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "replace into restore_log("
	  "cid,"
	  "backupid,"
	  "master_nodeid,"
	  "mgm_nodeid,"
	  "status"
	  ") values(%d,%d,%d,%d,'%s')",
	  g_clusterid,
	  backupid,
	  source_node,
	  node,
	  log ? "RestoreLog":"RestoreData");
  
  log_message(LOG_INFO, msg);
  
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  strcpy(restore_stats[node].status,log ? "RestoreLog":"RestoreData");
  //handler_restore_event(node,restore_stats[node]);  
  memset(restore_stats,0,256*sizeof(statistics_restore_t));      
}


void logger_restoremeta (int source_node, 
			int node, 
			int backupid)
{


  char msg[512];

  
  sprintf(msg,
	  "update restore "
	  "set "
	  "status='%s',"
	  "n_tables=%d,"
	  "n_tablespaces=%d,"
	  "n_logfilegroups=%d,"
	  "n_datafiles=%d,"
	  "n_undofiles=%d "
	  "where "
	  "cid=%d and backupid=%d and "
	  "ndb_nodeid=%d",	  
	  "RestoreMeta",
	  restore_stats[node].n_tables,
	  restore_stats[node].n_tablespaces,
	  restore_stats[node].n_logfilegroups,
	  restore_stats[node].n_datafiles,
	  restore_stats[node].n_undofiles,
	  g_clusterid,
	  backupid,
	  node);
  

  
  log_message(LOG_INFO, msg);
  
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "replace into restore_log("
	  "cid,"
	  "backupid,"
	  "master_nodeid,"
	  "mgm_nodeid,"
	  "status"
	  ") values(%d,%d,%d,%d,'%s')",
	  g_clusterid,
	  backupid,
	  source_node,
	  node,
	  "RestoreMeta");
  
  log_message(LOG_INFO, msg);
  
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  strcpy(restore_stats[node].status,"RestoreMeta");
  //handler_restore_event(node,restore_stats[node]);  
  memset(restore_stats,0,256*sizeof(statistics_restore_t));      
}



void logger_restore (int source_node, 
		    int node, 
		    int backupid, 
		    const char * status)
{
  
  char msg[512];
  
  
  if(strcmp(status,"RestoreStarted")==0)
    {
      sprintf(msg,
	      "replace into restore("
	      "cid,"
	      "backupid,"
	      "master_nodeid,"
	      "ndb_nodeid,"
	      "status"
	      ") values(%d,%d,%d,%d,'%s')",
	      g_clusterid,
	      backupid,
	      source_node,
	      node,
	      status);
      
      log_message(LOG_INFO, msg);
      executeQuery(msg,&mysql_conn[LOG_THREAD]);
      memset(msg,0,sizeof(msg));

    }
  else
    {
      sprintf(msg,
	      "update restore "
	      "set "
	      "status='%s' "
	      "where "
	      "backupid= %d and cid=%d and"
	      "ndb_nodeid=%d",	      
	      status,
	      backupid,
	      g_clusterid,
	      node);

      
      log_message(LOG_INFO, msg);
      executeQuery(msg,&mysql_conn[LOG_THREAD]);
    
    
    }
  
  memset(msg,0,sizeof(msg));
  

  sprintf(msg,
	  "replace into restore_log("
	  "cid,"
	  "backupid,"
	  "master_nodeid,"
	  "mgm_nodeid,"
	  "status"
	  ") values(%d,%d,%d,%d,'%s')",
	  g_clusterid,
	  backupid,
	  source_node,
	  node,
	  status);
  
  log_message(LOG_INFO, msg);
  
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  strcpy(restore_stats[node].status, status);
  //handler_restore_event(node,restore_stats[node]);  
  memset(restore_stats,0,256*sizeof(statistics_restore_t));      
  
}

void logger_backup (int source_node, 
		   int node)
{  
  char msg[512];
  sprintf(msg,
	  "replace into backup("
	  "cid,"
	  "backupid,"
	  "master_nodeid,"
	  "mgm_nodeid,"
	  "status,"
	  "start_gci,"
	  "stop_gci,"
	  "records,"
	  "log_records,"
	  "bytes,"
	  "log_bytes,"
	  "error"
	  ") values ("
	  "%d,%d,%d, %d, '%s', %d,%d,%llu,%llu,%llu,%llu,%d"
	  ")", 
	  g_clusterid,
	  backup_stats[node].backup_id,
	  source_node,
	  node,
	  backup_stats[node].status,
	  backup_stats[node].start_gci,
	  backup_stats[node].stop_gci,
	  backup_stats[node].n_records64,
	  backup_stats[node].n_log_records64,
	  backup_stats[node].n_bytes64,
	  backup_stats[node].n_log_bytes64,
	  backup_stats[node].error);	 
  

  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "insert into backup_log("
	  "cid,"
	  "backupid,"
	  "status,"
	  "error"
	  ") values ("
	  "%d, %d, '%s',%d"
	  ")", 
	  g_clusterid,
	  backup_stats[node].backup_id,
	  backup_stats[node].status,
	  backup_stats[node].error);

  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  //handler_backup_event(node,backup_stats[node]);  
  memset(backup_stats,0,256*sizeof(statistics_backup_t));      
}



void logger_memory ( int node)		   
{		 
  char msg[512];
  int dm_mb_used=(mem_stats[node].used_dm * mem_stats[node].page_sz_dm)/1024/1024;
  int im_mb_used=(mem_stats[node].used_im * mem_stats[node].page_sz_im)/1024/1024;
  int dm_mb_total=(mem_stats[node].total_dm * mem_stats[node].page_sz_dm)/1024/1024;
  int im_mb_total=(mem_stats[node].total_im * mem_stats[node].page_sz_im)/1024/1024;

  sprintf(msg,
	  "replace into memory_usage("
	  "cid,"
	  "nodeid,"
	  "pg_used_dm,"
	  "pg_tot_dm,"
	  "pg_used_im,"
	  "pg_tot_im,"
	  "pg_sz_dm,"
	  "pg_sz_im,"
	  "dm_used_mb,"
	  "im_used_mb,"
	  "dm_total_mb,"
	  "im_total_mb,"
	  "report_ts"
	  ") values ("
	  "%d,%d,%d, %d, %d, %d, %d, %d,%d,%d,%d,%d,now()"
	  ")", 
	  g_clusterid,
	  node,
	  mem_stats[node].used_dm,
	  mem_stats[node].total_dm,
	  mem_stats[node].used_im,
	  mem_stats[node].total_im,
	  mem_stats[node].page_sz_dm,
	  mem_stats[node].page_sz_im,
	  dm_mb_used,
	  im_mb_used,
	  dm_mb_total,
	  im_mb_total);
  
  //  handler_memory_usage_event(node,mem_stats[node]);
  
  /*
   * Check if we have an alam
   */


  executeQuery(msg,&mysql_conn[LOG_THREAD]);  
  memset(&mem_stats[node],0,sizeof(statistics_mem_t));
  check_mem_usage_alarm();

}

bool check_mem_usage_alarm()
{
  int percent_dm=0;
  int percent_im=0;
  //

  char msg[512];
  char query[128];
  snprintf(query,sizeof(query),"select 100*sum(pg_used_dm)/sum(pg_tot_dm), 100*sum(pg_used_im)/sum(pg_tot_im) as percent_im from memory_usage where cid=%d",g_clusterid);
  if(mysql_real_query(&mysql_conn[LOG_THREAD].mysql, query, strlen(query)))
    {
      snprintf(msg,sizeof(msg), "Query %s failed: Error: %s\n",
	      query,
	      mysql_error(&mysql_conn[LOG_THREAD].mysql));
      log_message(LOG_WARNING, msg);
      
      return false;
    }
  else // query succeeded, process any data returned by it
    {
      MYSQL_RES * result = mysql_store_result(&mysql_conn[LOG_THREAD].mysql);
      if (result)  // there are rows
	{
	  
	  MYSQL_ROW row;

	  unsigned long long num_rows = mysql_affected_rows(&mysql_conn[LOG_THREAD].mysql);
	  if(num_rows>0)
	    {
	      while ((row = mysql_fetch_row(result)))
		{
		  percent_dm=atoi(row[0]);
		  percent_im=atoi(row[1]);
		}
	    }	  
	  mysql_free_result(result);	 
	}
    }
  

  if(percent_dm>=80 && percent_dm<90)
    {
      char description[64];      
      snprintf(description,sizeof(description), "DataMemory usage is above %d percent",percent_dm);
      db_write_resource_alarm(&mysql_conn[LOG_THREAD],			   
			      0,
			      "WARNING",
			      "DATAMEMORY",
			      "RESOURCE ALARM",
			      "CLUSTER",
			      description,
			      "Alarm will be cleared when DataMemory drops below 80%.");
    }
  else if(percent_dm>=90)    
    {
      char description[64];      
      snprintf(description,sizeof(description), "DataMemory usage is above %d percent",percent_dm);

      db_write_resource_alarm(&mysql_conn[LOG_THREAD],			   
			      0,
			      "ALERT",
			      "DATAMEMORY",
			      "RESOURCE ALARM",
			      "CLUSTER",
			      description,
			      "Alarm will be cleared when DataMemory drops below 80%.");
      
    }
  //alarm
  else
    {
      //clear
      if(percent_dm<=80)
	{	  
	  db_clear_resource_alarm(&mysql_conn[LOG_THREAD], 0, "DATAMEMORY","IndexMemory usage <=80 percent."); 
	}
      
    }
  
  
  if(percent_im>=80 && percent_im<90)
    {
      char description[64];      
      snprintf(description,sizeof(description), "DataMemory usage is above %d percent",percent_dm);
      db_write_resource_alarm(&mysql_conn[LOG_THREAD],			   
			      0,
			      "WARNING",
			      "INDEXMEMORY",
			      "RESOURCE ALARM",
			      "CLUSTER",
			      description,
			      "Alarm will be cleared when IndexMemory drops below 80%.");
    }
  //alarm
  else if(percent_im>=90)    
    {
      char description[64];      
      snprintf(description,sizeof(description), "DataMemory usage is above %d percent",percent_dm);
      db_write_resource_alarm(&mysql_conn[LOG_THREAD],			   
			      0,
			      "ALERT",
			      "INDEXMEMORY",
			      "RESOURCE ALARM",
			      "CLUSTER",
			      description,
			      "Alarm will be cleared when IndexMemory drops below 80%.");
    }
  else
    {
      if(percent_im<=80)
	{
	  db_clear_resource_alarm(&mysql_conn[LOG_THREAD],0, "INDEXMEMORY", "IndexMemory usage <=80 percent.");		       
	}
    }

  return true;  
}


void * log_event_thread(void * t)
{
  
  log_message(LOG_INFO, "Starting LogEventThread");  

  char msg[512];
  int iterations = -1;
  if(!connectMySQL(&mysql_conn[LOG_THREAD]))
    exit(-1);
  init();
  disconnectMySQL(&mysql_conn[LOG_THREAD]);
  struct ndb_logevent event;
  NdbLogEventHandle le;
  while (true)
    {
      while(!connectNdbMgm(&cluster_conn[LOG_THREAD])) //check and connect to management server..
	{

	  usleep(1000*1000);
	}
      /*
      if(!connectMySQL(&mysql_conn[LOG_THREAD]))
	{
	  disconnectMySQL(&mysql_conn[LOG_THREAD]);
	  usleep(1000*1000);
          continue;
	}
      */
      le= ndb_mgm_create_logevent_handle(cluster_conn[LOG_THREAD].handle, filter);
      if(le==0)
	continue;
      int timeout= 50; //ms
      int timeout2= 1000; //ms
      
      while (iterations-- != 0)
	{
	  if(!connectMySQL(&mysql_conn[LOG_THREAD]))
	    {
	      disconnectMySQL(&mysql_conn[LOG_THREAD]);
	      usleep(1000*1000);
	      continue;
	    }
	  
	  int r= ndb_logevent_get_next(le,&event,timeout);
	  if (r == 0)
	    {
#ifdef DEBUG
	      printf("LogEventThread: No event within %d milliseconds\n", timeout);
#endif
	    }	  
	  else if (r < 0)
	    {

	      sprintf(msg, "LogEventThread: line: %d, code: %d msg: %s.", 
		      __LINE__,
		      ndb_logevent_get_latest_error(le), 
		      ndb_logevent_get_latest_error_msg(le));

	      log_message(LOG_INFO, msg);
	      ndb_mgm_destroy_logevent_handle(&le);	      
	      while(!connectNdbMgm(&cluster_conn[LOG_THREAD]))
		{
		  usleep(timeout2*1000);		  
		}      
	      if(ndb_mgm_is_connected(cluster_conn[LOG_THREAD].handle)>0)
		{
		  le=0;
		  while(le==0)
		    {
		      le= ndb_mgm_create_logevent_handle(cluster_conn[LOG_THREAD].handle, 
							 filter);
		      usleep(timeout*1000);
		    }

		}
	      continue;
	    }
	  else
	    {

	      memset(msg,0,sizeof(msg));
	      switch (event.type) {
	      case NDB_LE_NDBStopForced:
		/*
		 * Here we need to check what error code that StopForced returned
		 * and set if we should do a Initial Node Restart or a Node Restart.
		 */
		{
		  sprintf(msg, 
			  "Node %d: StopForced"
			  " Error: %d", 
			  event.source_nodeid,
			  event.NDBStopForced.error);
		  
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "StopForced",
			    msg);

		  log_message(LOG_ALERT,msg);
		  if(cluster_state.nodes[event.source_nodeid].type ==
		     NDB_MGM_NODE_TYPE_NDB)
		    {
		      ndbd_exit_classification nexl;
		      ndbd_exit_status  nes;
		      
		      const char * exit_desc=ndbd_exit_message(event.NDBStopForced.error, &nexl);
		      const char * class_desc=ndbd_exit_classification_message(nexl,&nes);
		      const char * desc = ndbd_exit_status_message(nes);
		      char reco[512];
		      strcat(reco, desc);
		      strcat(reco, ". Wait for COMMUNICATION OPENED alarm to come before restarting node!");
     		      if(db_check_node_alarm(  event.source_nodeid, &mysql_conn[LOG_THREAD]))
			{			 
			  db_update_alarm_error(&mysql_conn[LOG_THREAD],  
						event.source_nodeid,
						exit_desc,
						class_desc,
						reco);		
			}
		      else
			{
			  db_write_alarm(&mysql_conn[LOG_THREAD],			   
					 g_clusterid,
					 event.source_nodeid,
					 cluster_state.current_state,
					 cluster_state.current_state==CS_DEGRADED ? CS_DEGRADED: cluster_state.previous_state ,
					 "ALERT",
					 "NDBD",
					 "NODE DISCONNECTED",
					 "No error code exists to map error message.",
					 "Data node disconnected.",
					 "Restart data node! Wait for COMMUNICATION_OPENED ALARM before restarting node!"
					 );
			}
		      

		    }

		  cluster_state.nodes[event.source_nodeid].startmode=CS_INR;
		}
		
		

		break;
	      case NDB_LE_NDBStartCompleted:
		/*
		 * The node has started.
		 * The default way to recover it is Node Recovery (CS_NR);
		 * cluster_state.nodes[event.Disconnected.node].startIsOk=false;
		 * means that it is not ok to try to start a STARTED node
		 */
		{
		  
		  
		  cluster_state.nodes[event.source_nodeid].startmode=CS_NR;
		  cluster_state.nodes[event.source_nodeid].startIsOk=false;
		  
		  sprintf(msg, 
			  "Node %d: Start Completed "
			  "version %d.%d.%d",
			  event.source_nodeid,
			  event.NDBStartCompleted.version >> 16 & 0xff,
			  event.NDBStartCompleted.version >> 8 & 0xff,
			  event.NDBStartCompleted.version >> 0 & 0xff);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "StartCompleted",
			    msg);		 
		  if(cluster_state.nodes[event.source_nodeid].type ==
		     NDB_MGM_NODE_TYPE_NDB)
		    {		    
		      db_node_clear_alarm( &mysql_conn[LOG_THREAD],		      
				           event.source_nodeid,
					   "NDBD",
					   "ALERT",
					   "NODE DISCONNECTED");
		    }
		  

		}

		break;
	      case NDB_LE_CommunicationOpened:
		{
		  /*
		   * cluster_state.nodes[event.Disconnected.node].startIsOk=true;
		   * means that it is  ok to try to start a STARTED node
		   * We should only try to START a node when we have got the 
		   * CommmunicationOpened, since we then have seen
		   * the NdbStopForced log event (if it has arrived) telling
		   * us HOW to restart the node (INR or NR).
		   */
		  cluster_state.nodes[event.CommunicationOpened.node].startIsOk=true;
		  
		  
		  sprintf(msg, 
			  "Node %d: Communication opened to node %d", 
			  event.source_nodeid,
			  event.CommunicationOpened.node);
		  
		  logger_db (event.source_nodeid, 
			    event.CommunicationOpened.node, 
			    event.severity,event.level,
			     "CommunicationOpened",
			     msg);

		  if(cluster_state.nodes[event.CommunicationOpened.node].type ==
		     NDB_MGM_NODE_TYPE_NDB)
		    {
     		      if(db_check_node_alarm(event.CommunicationOpened.node, &mysql_conn[LOG_THREAD]))
			{			
			  db_write_alarm(&mysql_conn[LOG_THREAD],			   
					 g_clusterid,
					 event.CommunicationOpened.node,
					 CS_DEGRADED,
					 cluster_state.current_state==CS_DEGRADED ? CS_DEGRADED: cluster_state.previous_state ,
					 "INFO",
					 "NDBD",
					 "COMMUNICATION OPENED",					 
					 "",
					 "Restart of failed data node is now possible.",
					 "Restart data node!");
			  db_update_alarm_count(&mysql_conn[LOG_THREAD],  
						event.CommunicationOpened.node);

			}
		    }
		  
		}
		break;
	      case NDB_LE_CommunicationClosed:
		{
		  /*
		   * cluster_state.nodes[event.Disconnected.node].startIsOk=false;
		   * means that it is not ok to try to start a STARTED node
		   */
		  cluster_state.nodes[event.CommunicationClosed.node].startIsOk=false;
		  
		  sprintf(msg, 
			"Node %d: Communication closed to node %d", 
			  event.source_nodeid,
			  event.CommunicationClosed.node);

		  logger_db (event.source_nodeid, 
			    event.CommunicationClosed.node, 
			    event.severity,event.level,
			    "CommunicationClosed",
			    msg);

		}
		  
		break;
	      case NDB_LE_Disconnected:
		/*
		 * cluster_state.nodes[event.Disconnected.node].startIsOk=false;
		 * means that it is not ok to try to start a STARTED node
		 */
		{
		  cluster_state.nodes[event.Disconnected.node].startIsOk=false;
		  
		  
		  sprintf(msg, "Node %d: Disconnected from node %d", 
			  event.source_nodeid,
			  event.Disconnected.node);

		  log_message(LOG_WARNING,msg);
		  
		  logger_db (event.source_nodeid, 
			    event.Disconnected.node, 
			    event.severity,event.level,
			    "Disconnected",
			    msg);
		  
		  if(cluster_state.nodes[event.Disconnected.node].type ==
		     NDB_MGM_NODE_TYPE_NDB)
		    {
		      memset(msg,0,strlen(msg));
		      sprintf(msg, 
			      "UPDATE node_state "
			      "SET disconnects=disconnects+1," 
			      "last_disconnect=NOW() WHERE nodeid=%d and cid=%d",
			      event.Disconnected.node,
			      g_clusterid);
		      executeQuery(msg,&mysql_conn[LOG_THREAD]);
		      db_write_alarm(&mysql_conn[LOG_THREAD],			   
				     g_clusterid,
				     event.Disconnected.node,
				     CS_DEGRADED,
				     cluster_state.current_state==CS_DEGRADED ? CS_DEGRADED: cluster_state.previous_state  ,		  
				     "ALERT",
				     "NDBD",
				     "NODE DISCONNECTED",
				     "No error code exists to map error message.",
				     "Data node disconnected.",
				     "Restart data node!."
				     );


		    }

		  if(cluster_state.nodes[event.Disconnected.node].type ==
		     NDB_MGM_NODE_TYPE_API)
		    {
		      if(is_mysqld(event.Disconnected.node,&mysql_conn[LOG_THREAD]))
			{
			  db_write_alarm(&mysql_conn[LOG_THREAD],			   
					 g_clusterid,
					 event.Disconnected.node,
					 CS_DEGRADED,
					 cluster_state.current_state==CS_DEGRADED ? CS_DEGRADED: cluster_state.previous_state  ,       
					 "ALERT",
					 "MYSQLD",
					 "MYSQL SERVER DISCONNECTED",
					 "MySQL server disconnected",
					 "Check error log of failed mysql server",
					 "Restart mysql server");
			}
		    }
		}
		break;

	      case NDB_LE_Connected:
		/*
		 * cluster_state.nodes[event.Connected.node].startIsOk=false;
		 * means that it is not ok to try to start a STARTED node
		 */
		{
		  if(cluster_state.nodes[event.Connected.node].type ==
		     NDB_MGM_NODE_TYPE_API)
		    {
  		      if(is_mysqld(event.Connected.node,&mysql_conn[LOG_THREAD]))
			{
			  db_node_clear_alarm( &mysql_conn[LOG_THREAD],		      
					       event.Connected.node,
					       "MYSQLD",
					       "ALERT",
					       "MYSQL SERVER DISCONNECTED");
					       
			}
		    }
		  if(  event.Connected.node !=0)
		    {
		      sprintf(msg, "Node %d: Connected to node %d", 
			      event.source_nodeid,
			      event.Connected.node);
		      
		      
		      
		      logger_db (event.source_nodeid, 
				 event.Disconnected.node, 
				 event.severity,event.level,
				 "Connected",
				 msg);
		    }
		}
		break;
		
	      case NDB_LE_StartPhaseCompleted:
		{
		  
		  
		  sprintf(msg, 
			  "Node %d: StartPhase completed"
			  " phase = %d type  = %d", 
			  event.source_nodeid, 
			  event.StartPhaseCompleted.phase, 
			  event.StartPhaseCompleted.starttype);
		  
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "StartPhaseCompleted",
			    msg);
		}
		
		break;

	      case NDB_LE_NDBStartStarted:
		{
		  
		  sprintf(msg, 
			  "Node %d: StartStarted "
			  "version %d.%d.%d",
			  event.source_nodeid,		 
			  event.NDBStartStarted.version >> 16 & 0xff,
			  event.NDBStartStarted.version >> 8 & 0xff,
			  event.NDBStartStarted.version >> 0 & 0xff);
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "StartStarted",
			    msg);
		}
		break;

	      case NDB_LE_NDBStopStarted:
		{
		  
		  sprintf(msg, 
			  "LogEventThread: Node %d: StopStarted "
			  "type  = %d", 
			  event.source_nodeid, 
			  event.NDBStopStarted.stoptype); 
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "StopStarted",
			    msg);
		}

		break;

	      case NDB_LE_NDBStopCompleted:
		{
		  
		  sprintf(msg, "Node %d: StopCompleted", 
			  event.source_nodeid);
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "StopCompleted",
			    msg);
		}
		
		break;
		
	      case NDB_LE_ConnectedApiVersion:
		{
		  
		  sprintf(msg,
			  "Node %d: ConnectedApiVersion "
			  "connected node %d , version %d.%d.%d",
			  event.source_nodeid,
			  event.ConnectedApiVersion.node, 
			  event.ConnectedApiVersion.version >> 16 & 0xff,
			  event.ConnectedApiVersion.version >> 8 & 0xff,
			  event.ConnectedApiVersion.version >> 0 & 0xff);
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "ConnectedApiVersion",
			    msg);
		}
		break;
	
	      case NDB_LE_BackupStarted:
		{
		  
		  
		  sprintf(msg, 
			  "Node %d: Backup started by node %d "
			  " and the Backup ID = %d",
			  event.source_nodeid,
			  event.BackupStarted.starting_node,
			  event.BackupStarted.backup_id);
		  logger_db (event.source_nodeid, 
			    event.BackupStarted.starting_node,
			    event.severity,event.level,
			    "BackupStarted",
			    msg);
		  
		  backup_stats[event.BackupStarted.starting_node].backup_id=
		    event.BackupStarted.backup_id,
		  strcpy(backup_stats[event.BackupStarted.starting_node].status,
			 "BackupStarted");			      
		  logger_backup (      
				event.source_nodeid,
				event.BackupStarted.starting_node);
		  
		}
		break;
	      case NDB_LE_BackupCompleted:
		{
		  
		  
		  sprintf(msg, 
			  "Node %d: Backup completed. "
			  "Backup ID = %d",
			  event.source_nodeid,
			  event.BackupCompleted.backup_id);
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "BackupCompleted",
			    msg);
		  
		  strcpy(backup_stats[event.BackupCompleted.starting_node].status,
					 "BackupCompleted");			      
		  backup_stats[event.BackupCompleted.starting_node].start_gci
		    =event.BackupCompleted.start_gci;
		  
		  backup_stats[event.BackupCompleted.starting_node].stop_gci =event.BackupCompleted.stop_gci;
		  
		  backup_stats[event.BackupCompleted.starting_node].n_bytes64=
		    ((unsigned long long)(event.BackupCompleted.n_bytes_hi)<< 32) | 
		    event.BackupCompleted.n_bytes;
		  backup_stats[event.BackupCompleted.starting_node].n_log_bytes64=
		    ((unsigned long long)(event.BackupCompleted.n_log_bytes_hi)<< 32)
		    | event.BackupCompleted.n_log_bytes;
		  backup_stats[event.BackupCompleted.starting_node].n_records64=
		    ((unsigned long long)event.BackupCompleted.n_records_hi) << 32 | 
		    event.BackupCompleted.n_records;
		  
		  backup_stats[event.BackupCompleted.starting_node].n_log_records64=
		    ((unsigned long long)(event.BackupCompleted.n_log_records_hi) 
		     << 32) | event.BackupCompleted.n_log_records;

		  backup_stats[event.BackupCompleted.starting_node].backup_id=
		    event.BackupCompleted.backup_id;

		  logger_backup (event.source_nodeid,
				event.BackupCompleted.starting_node);
		  
		  
		}
		break;
	      case NDB_LE_BackupStatus:
		{
		  		
		  strcpy(backup_stats[event.BackupCompleted.starting_node].status,
			 "BackupStatus");			        
		  backup_stats[event.BackupStatus.starting_node].backup_id=
		    event.BackupStatus.backup_id;
		  logger_backup (      
				event.source_nodeid,
				event.BackupStatus.starting_node);		  
		  
		}
		break;
	      case NDB_LE_BackupAborted:
		{
		  
		  
		  sprintf(msg, 
			  "Node %d: Backup aborted  by node %d. "
			  "Backup ID = %d",
			  event.source_nodeid,
			  event.BackupAborted.starting_node,
			  event.BackupAborted.backup_id);
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "BackupAborted",
			    msg);
		  strcpy(backup_stats[event.BackupAborted.starting_node].status,
			 "BackupAborted");			        
		  backup_stats[event.BackupAborted.starting_node].backup_id=
		    event.BackupAborted.backup_id;
		  backup_stats[event.BackupAborted.starting_node].error=
		    event.BackupAborted.error;
		  logger_backup (      
				event.source_nodeid,
				event.BackupAborted.starting_node);
		  
		}
		break;

	      case NDB_LE_BackupFailedToStart:
		{
		  
		  
		  sprintf(msg, 
			  "Node %d: Backup failed to start.  by node %d. "
			  "Error = %d",
			  event.source_nodeid,
			  event.BackupFailedToStart.starting_node,
			  event.BackupFailedToStart.error);
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "BackupFailedToStart",
			    msg);
		  strcpy(backup_stats[event.BackupFailedToStart.starting_node].status,
			 "BackupFailedToStart");			        
		  backup_stats[event.BackupFailedToStart.starting_node].backup_id=0;
		  backup_stats[event.BackupFailedToStart.starting_node].error=
		    event.BackupFailedToStart.error;
		  
		  logger_backup (event.source_nodeid,
				event.BackupFailedToStart.starting_node);
		  
		  
		}
		break;

	      case NDB_LE_RestoreMetaData:
		{
		  restore_stats[event.RestoreMetaData.node_id].n_tables=
		    event.RestoreMetaData.n_tables;
		  restore_stats[event.RestoreMetaData.node_id].n_tablespaces=
		    event.RestoreMetaData.n_tablespaces;
		  restore_stats[event.RestoreMetaData.node_id].n_logfilegroups=
		    event.RestoreMetaData.n_logfilegroups;
		  restore_stats[event.RestoreMetaData.node_id].n_datafiles=
		    event.RestoreMetaData.n_datafiles;
		  restore_stats[event.RestoreMetaData.node_id].n_undofiles=
		    event.RestoreMetaData.n_undofiles;
		  logger_restoremeta (event.source_nodeid, 
				     event.RestoreMetaData.node_id,
				     event.RestoreMetaData.backup_id);
		}
		break;
	      case NDB_LE_RestoreData:
		{
		  backup_stats[event.RestoreData.node_id].n_bytes64=
		    ((unsigned long long)(event.RestoreData.n_bytes_hi)<< 32)
		    | event.RestoreData.n_bytes_lo;
		  backup_stats[event.RestoreData.node_id].n_records64=
		    ((unsigned long long)(event.RestoreData.n_records_hi) 
			       << 32) |  event.RestoreData.n_records_lo;

		  logger_restore_datalog (event.source_nodeid, 
					event.RestoreData.node_id,
					event.RestoreData.backup_id,
					false);
		  sprintf(msg, 
			  "Node %d: Restore Data of Backup ID %d - "
			  "Restore was started by node %d. Restored %llu bytes",
			  event.source_nodeid, 
			  event.RestoreData.backup_id,
			  event.RestoreData.node_id,
			  backup_stats[event.RestoreData.node_id].n_bytes64 );

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "RestoreData",
			    msg);
		  
		    }
		break;
	      case NDB_LE_RestoreLog:
		{
		  backup_stats[event.RestoreData.node_id].n_bytes64=
		    ((unsigned long long)(event.RestoreData.n_bytes_hi)<< 32)
		    | event.RestoreData.n_bytes_lo;
		  backup_stats[event.RestoreData.node_id].n_records64=
		    ((unsigned long long)(event.RestoreData.n_records_hi) 
			       << 32) |  event.RestoreData.n_records_lo;

		  logger_restore_datalog (event.source_nodeid, 
					event.RestoreData.node_id,
					event.RestoreData.backup_id,
					false);
		  sprintf(msg, 
			  "Node %d: Restore Log of Backup ID %d - "
			  "Restore was started by node %d. "
			  "Restored %llu bytes of log",
			  event.source_nodeid, 
			  event.RestoreData.backup_id,
			  event.RestoreData.node_id,
			  backup_stats[event.RestoreData.node_id].n_bytes64 );

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "RestoreData",
			    msg);
		}
		break;
	      case NDB_LE_RestoreStarted:
		{
		  logger_restore (event.source_nodeid, 
				 event.RestoreStarted.node_id,
				 event.RestoreStarted.backup_id,
				 "RestoreStarted"
				 );
		  sprintf(msg, 
			  "Node %d: Restore Started of Backup ID %d - "
			  "Restore was started by node %d. ",
			  event.source_nodeid, 
			  event.RestoreStarted.backup_id,
			  event.RestoreStarted.node_id);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "RestoreStarted",
			    msg);
				
		}
		break;
	      case NDB_LE_RestoreCompleted:
		{
		  logger_restore (event.source_nodeid, 
				 event.RestoreCompleted.node_id,
				 event.RestoreCompleted.backup_id,
				 "RestoreCompleted"
				 );

		  sprintf(msg, 
			  "Node %d: Restore Completed of Backup ID %d - "
			  "Restore was started by node %d. ",
			  event.source_nodeid, 
			  event.RestoreCompleted.backup_id,
			  event.RestoreCompleted.node_id);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "RestoreCompleted",
			    msg);

		}
		break;

	      case NDB_LE_NodeFailCompleted:
		{
		  sprintf(msg, "Node %d: NodeFailCompleted of node %d"
			  " in block %d",
			  event.source_nodeid,
			  event.NodeFailCompleted.failed_node,
			  event.NodeFailCompleted.block);
		  logger_db (event.source_nodeid, 
			    event.NodeFailCompleted.failed_node,
			    event.severity,event.level,
			    "NodeFailCompleted",
			    msg);

		}
		break;
	      case NDB_LE_ArbitResult:
		{
		  
		  
		  sprintf(msg, "Node %d: ArbitResult "
			  "code %d, arbit_node %d",
			  event.source_nodeid,
			  event.ArbitResult.code & 0xffff,
			  event.ArbitResult.arbit_node);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "ArbitResult",
			    msg);
		}
		break;
	      case NDB_LE_MissedHeartbeat:
		{
		  
		  sprintf(msg, 
			  "Node %d: Missed Heartbeat->nodeid=%d", 
			  event.source_nodeid,
			  event.MissedHeartbeat.node);
		  logger_db (event.source_nodeid, 
			    event.MissedHeartbeat.node,
			    event.severity,event.level,
			    "MissedHeartbet",
			    msg);
		}
		break;
	      case NDB_LE_DeadDueToHeartbeat:
		{
		  
		  sprintf(msg, 
			  "Node %d: DeadDueToHeartbeat->nodeid=%d", 
			  event.source_nodeid,
			  event.DeadDueToHeartbeat.node);
		  logger_db (event.source_nodeid, 
			    event.DeadDueToHeartbeat.node,
			    event.severity,event.level,
			    "DeadDueToHeartbeat",
			    msg);
		}
		break;
	      case NDB_LE_ArbitState:
		{
		  
		  
		  sprintf(msg, 
			  "Node %d: ArbitResult "
			  "code %d, arbit_node %d",
			  event.source_nodeid,
			  event.ArbitState.code & 0xffff,
			  event.ArbitState.arbit_node);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "ArbitState",
			    msg);
		}
		break;
	      case NDB_LE_NR_CopyFragsStarted:
		{
		  
		  sprintf(msg,
			  "Node %d: Starting Copying Fragments"
			  " (CopyFragsStarted) to node %d",
			  event.source_nodeid,
			  event.NR_CopyFragsStarted.dest_node);

		  logger_db (event.source_nodeid, 
			    event.NR_CopyFragsStarted.dest_node,
			    event.severity,event.level,
			    "CopyFragsStared",
			    msg);
		}
		
		break;
	      case NDB_LE_NR_CopyDict:
		{
		  
		  sprintf(msg, 
			  "Node %d: Copying Dictionary information",
			  event.source_nodeid);
		  
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "CopyDict",
			    msg);
		}
		
		break;

	      case NDB_LE_NR_CopyDistr:
		{
		  
		  sprintf(msg, 
			  "Node %d: Copying Distribution information",
			  event.source_nodeid);
		  
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "CopyDistr",
			    msg);
		}
		
		break;
		
	      case NDB_LE_NR_CopyFragDone:
		{
		  
		  sprintf(msg, "Node %d: Copy Fragment (CopyFragDone) "
			  "to node %d"
			" (frag id = %d, "	       
			  "tableid=%d) is DONE",
			  event.source_nodeid,
			  event.NR_CopyFragDone.dest_node,
			  event.NR_CopyFragDone.fragment_id,
			  event.NR_CopyFragDone.table_id);

		  logger_db (event.source_nodeid, 
			    event.NR_CopyFragDone.dest_node,
			    event.severity,event.level,
			    "CopyFragDone",
			    msg);
		  
		}
		break;

	      case NDB_LE_NR_CopyFragsCompleted:
		{
		  
		  sprintf(msg, "Node %d: Completed Copying Fragments"
			  " (CopyFragsCompleted) to node %d",
			  event.source_nodeid,
			  event.NR_CopyFragsStarted.dest_node);

		  logger_db (event.source_nodeid, 
			    event.NR_CopyFragsCompleted.dest_node,
			    event.severity,event.level,
			    "CopyFragsCompleted",
			    msg);
		}
		break;
	      case NDB_LE_UNDORecordsExecuted:
		{
		  

		  sprintf(msg, 
			  "Node %d: Executing Undo Records",
			  event.source_nodeid);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "UNDORecordsExecuted",
			    msg);
		}
		break;
	      case NDB_LE_STTORRYRecieved:
		{
		  /*
		  sprintf(msg, "STTORRYReceived...nothing to handle");
		  */
		}
		break;
	      case NDB_LE_CM_REGCONF:
		{
		  
		  sprintf(msg, 
			  "Node %d: CM_REGCONF. Inclusion in cluster ok: "
			  "my id=%d, president id =%d"
			  ", dynamic id=%d",
			  event.source_nodeid,
			  event.CM_REGCONF.own_id,
			  event.CM_REGCONF.president_id,
			  event.CM_REGCONF.dynamic_id);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "CM_REGCONF",
			    msg);
		  
		}
		break;
	      case NDB_LE_CM_REGREF:
		{
		  
		  sprintf(msg, 
			  "Node %d: CM_REGREF. Inclusion in cluster failed: "
			  "my id=%d, other id =%d"
			  ", cause=%d",
			  event.source_nodeid,
			  event.CM_REGREF.own_id,
			  event.CM_REGREF.other_id,
			  event.CM_REGREF.cause);
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "CM_REGREF",
			    msg);
		  
		}
		break;
	      case NDB_LE_FIND_NEIGHBOURS:
		{
		  
		  sprintf(msg, "Node %d: FIND_NEIGHBOURS. My neighbours: "
			  "my id=%d, left node id=%d,  right node id=%d"
			  ", dynamic id=%d",
			  event.source_nodeid,
			  event.FIND_NEIGHBOURS.own_id,
			  event.FIND_NEIGHBOURS.left_id,
			  event.FIND_NEIGHBOURS.right_id,
			  event.FIND_NEIGHBOURS.dynamic_id);
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "FIND_NEIGHBOURS",
			    msg);
		}
		break;
	      case NDB_LE_NODE_FAILREP:
		{
		  
		  sprintf(msg, "Node %d: Node failure reported. "
			  " Nodeid %d failed, state %d",
			  event.source_nodeid,
			  event.NODE_FAILREP.failed_node,
			  event.NODE_FAILREP.failure_state);
		  log_message(LOG_ALERT,msg);

		  logger_db (event.source_nodeid, 
			    event.NODE_FAILREP.failed_node,
			    event.severity,event.level,
			    "NODE_FAILREP",
			    msg);

		}
		break;
	      case NDB_LE_StartREDOLog:
		{
		  
		  sprintf(msg, "Node %d: Starting REDO LOG: "
			  "Keep GCI=%d, completed GCI=%d, restorable GCI=%d",
			  event.StartREDOLog.node,
			  event.StartREDOLog.keep_gci,
			  event.StartREDOLog.completed_gci,
			  event.StartREDOLog.restorable_gci);
		  
		  logger_db (event.source_nodeid, 
			    event.StartREDOLog.node,
			    event.severity,event.level,
			    "StartREDOLog",
			    msg);

		}
		break;
	      case NDB_LE_StartLog:
		{
		  
		  sprintf(msg, 
			  "Node %d: Starting LOG: "
			  "log part=%d, start mb=%d, stop mb=%d, GCI=%d",
			  event.source_nodeid,
			  event.StartLog.log_part,
			  event.StartLog.start_mb,
			  event.StartLog.stop_mb,
			  event.StartLog.gci);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "StartLog",
			    msg);

		}
		break;
	      case NDB_LE_GCP_TakeoverStarted:
		{
		  
		  sprintf(msg, 
			  "Node %d: GCP take over started",
			  event.source_nodeid);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "GCP_TakeoverStarted",
			    msg);
		}
		break;
	      case NDB_LE_GCP_TakeoverCompleted:
		{
		  
		  sprintf(msg, 
			  "Node %d: GCP take over completed\n ",
			  event.source_nodeid);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "GCP_TakeoverCompleted",
			    msg);
		}
		break;
	      case NDB_LE_LCP_TakeoverCompleted:
		{
		  
		  sprintf(msg, "Node %d: LCP take over completed, state=%d\n ",
			  event.source_nodeid,
			  event.LCP_TakeoverCompleted.state);

		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "LCP_TakeoverCompleted",
			    msg);
		}
		break;
	      case NDB_LE_LocalCheckpointStarted:
		{
		  
		  sprintf(msg, "Node %d: Local checkpoint %d started. Keep GCI = %d oldest restorable GCI = %d.\n ",
			  event.source_nodeid,
			  event.LocalCheckpointStarted.lci,
			  event.LocalCheckpointStarted.keep_gci,
			  event.LocalCheckpointStarted.restore_gci);
			  
		  
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "LocalCheckpointStarted",
			    msg);
		  
		}
		break;
	      case NDB_LE_LocalCheckpointCompleted:
		{
		  
		  sprintf(msg, "Node %d: Local checkpoint %d completed.(\n ",			  
			  event.source_nodeid,
			  event.LocalCheckpointCompleted.lci
			  );
		  
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "LocalCheckpointCompleted",
			    msg);
		  
		}
		break;
	      case NDB_LE_LCP_TakeoverStarted:
		{
		  
		  sprintf(msg, "Node %d: LCP take over started\n ",
			  event.source_nodeid);
		  
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "LCP_TakeoverStarted",
			    msg);
		  
		}
		break;
	      case NDB_LE_MemoryUsage:
		{
		  
		  if( event.MemoryUsage.block ==  248)
		    {
		      mem_stats[event.source_nodeid].used_im=
			event.MemoryUsage.pages_used;
		      mem_stats[event.source_nodeid].total_im=
			event.MemoryUsage.pages_total;
		      mem_stats[event.source_nodeid].page_sz_im=
			event.MemoryUsage.page_size_bytes;
		      sprintf(msg, 
			      "Node %d: Memory Usage (IndexMemory) used=%d, total=%d, page_size=%d\n",
			      event.source_nodeid,
			      event.MemoryUsage.pages_used,
			      event.MemoryUsage.pages_total,
			      event.MemoryUsage.page_size_bytes);
		    }
		  else
		    {
		      mem_stats[event.source_nodeid].used_dm=
			event.MemoryUsage.pages_used;
		      mem_stats[event.source_nodeid].total_dm=
			event.MemoryUsage.pages_total;
		      mem_stats[event.source_nodeid].page_sz_dm=
			event.MemoryUsage.page_size_bytes;
		      /*  sprintf(msg, 
			      "Node %d: Memory Usage (DataMemory) used=%d, total=%d, page_size=%d\n",
			      event.source_nodeid,
			      event.MemoryUsage.pages_used,
			      event.MemoryUsage.pages_total,
			      event.MemoryUsage.page_size_bytes);		      */
		    }
		  /*
		    logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "MemoryUsage",
			    msg);
		  */
		  if( mem_stats[event.source_nodeid].used_dm &&     
		      mem_stats[event.source_nodeid].total_dm &&
		      mem_stats[event.source_nodeid].used_im &&
		      mem_stats[event.source_nodeid].total_im)
		    {
		      logger_memory(event.source_nodeid);		      
		    }
		  
		}
		break;
	      case  NDB_LE_TransReportCounters:
		{
		  


		  report_tc_statistics(
				    event.source_nodeid,
				    (unsigned int)event.TransReportCounters.trans_count,
				    (unsigned int)event.TransReportCounters.write_count,
				    (unsigned int)event.TransReportCounters.commit_count,
				    (unsigned int)event.TransReportCounters.abort_count,
				    (unsigned int)event.TransReportCounters.attrinfo_count,
				    (unsigned int)event.TransReportCounters.read_count,
				    (unsigned int)event.TransReportCounters.simple_read_count,
				    (unsigned int)event.TransReportCounters.scan_count,
				    (unsigned int)event.TransReportCounters.range_scan_count,
				    (unsigned int)event.TransReportCounters.conc_op_count);
		}
		break;
	      case  NDB_LE_OperationReportCounters:
		{
		  report_lqh_statistics(event.source_nodeid,
					(unsigned int)event.OperationReportCounters.ops);
		}	       
		break;
	      case  NDB_LE_StartReport:
		{
		  printf("reporting node=%d report_type=%d, rem-time=%d, bmsize=%d, data=%d\n",
			 event.source_nodeid,			 
			 event.StartReport.report_type,
			 event.StartReport.remaining_time,
			 event.StartReport.bitmask_size,
			 event.StartReport.bitmask_data[0]);
		}
		break;
	      case NDB_LE_SubscriptionStatus:
		{
		  printf("reporting node=%d  report_type=%d, nodeid=%d\n",			 
			 event.source_nodeid,			 
			 event.SubscriptionStatus.report_type,
			 event.SubscriptionStatus.node_id);	     		  
		}
		break;
	      case NDB_LE_TableCreated:
		{

		  sprintf(msg, "Node %d: Table created id=%d\n ",
			  event.source_nodeid,
			  event.TableCreated.table_id);
		  
		  logger_db (event.source_nodeid, 
			    0,
			    event.severity,event.level,
			    "TableCreated",
			    msg);

		  logger_create_table (event.TableCreated.table_id);
		}
		break;
	      default:
		break;
	      }
	    }
	  disconnectMySQL(&mysql_conn[LOG_THREAD]);
	}
      //      disconnectMySQL(&mysql_conn[LOG_THREAD]);
    }
  ndb_mgm_destroy_logevent_handle(&le);
    
  free(perf_stats);
  free(mem_stats);
  

  return 0;
}

void report_lqh_statistics( unsigned int nodeid, 
			    unsigned int ops )
{
  char msg[2048];
  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "update node_statistics set "
	  "operations=operations + %d "
	  "where nodeid=%d and cid=%d",
	  ops,
	  nodeid,g_clusterid);
  executeQuery(msg,&mysql_conn[LOG_THREAD]);


  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "update cluster_statistics set "
	  "operations=operations + %d "
	  "where id=%d",
	  ops,
	  g_clusterid);
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  
}

void report_tc_statistics( unsigned int nodeid, 
			   unsigned int tc, 
			   unsigned int wc,
			   unsigned int cc,
			   unsigned int ac,
			   unsigned int aic,
			   unsigned int rc,
			   unsigned int src,
			   unsigned int sc,
			   unsigned int rsc,
			   unsigned int ccc)
{
  char msg[2048];
  
  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "update node_statistics set "
	  "transactions=transactions + %d," 
	  "commits=commits + %d,"
	  "writes=writes + %d," 
	  "aborts=aborts + %d,"
	  "attrs=attrs + %d,"
	  "reads_lock=reads_lock+%d,"
	  "reads_nolock=reads_nolock + %d,"
	  "scans=scans + %d,"
	  "range_scans = range_scans + %d,"
	  "conc_ops = conc_ops + %d,"
	  "report_ts=now()"
	  "where nodeid=%d and cid=%d",
	  tc,
	  cc,
	  wc,
	  ac,
	  aic,
	  rc,
	  src,
	  sc,
	  rsc,
	  ccc,
	  nodeid,
	  g_clusterid);
	  
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
  //log_message(LOG_INFO, msg);
  memset(msg,0,sizeof(msg));
  sprintf(msg,
	  "update cluster_statistics set "
	  "transactions=transactions + %d," 
	  "commits=commits + %d,"
	  "writes=writes + %d," 
	  "aborts=aborts + %d,"
	  "attrs=attrs + %d,"
	  "reads_lock=reads_lock+%d,"
	  "reads_nolock=reads_nolock + %d,"
	  "scans=scans + %d,"
	  "range_scans = range_scans + %d,"
	  "conc_ops = conc_ops + %d,"
	  "report_ts=now()"
	  "where id=%d",
	  tc,
	  cc,
	  wc,
	  ac,
	  aic,
	  rc,
	  src,
	  sc,
	  rsc,
	  ccc,
	  g_clusterid);
  executeQuery(msg,&mysql_conn[LOG_THREAD]);
}
